/* Distributed Checksum Clearinghouse checksum inverter
 *
 * report the previously computed checksums of a message
 *
 * Copyright (c) 2005 by Rhyolite Software, LLC
 *
 * This agreement is not applicable to any entity which sells anti-spam
 * solutions to others or provides an anti-spam solution as part of a
 * security solution sold to other entities, or to a private network
 * which employs the DCC or uses data provided by operation of the DCC
 * but does not provide corresponding data to other users.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * Parties not eligible to receive a license under this agreement can
 * obtain a commercial license to use DCC and permission to use
 * U.S. Patent 6,330,590 by contacting Commtouch at http://www.commtouch.com/
 * or by email to nospam@commtouch.com.
 *
 * A commercial license would be for Distributed Checksum and Reputation
 * Clearinghouse software.  That software includes additional features.  This
 * free license for Distributed ChecksumClearinghouse Software does not in any
 * way grant permision to use Distributed Checksum and Reputation Clearinghouse
 * software
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE, LLC DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE, LLC
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.3.29-1.4 $Revision$
 */

#include "dcc_ck.h"
#include "dcc_xhdr.h"
#include <arpa/inet.h>


static DCC_EMSG dcc_emsg;

static int verbose;
#define LOOP_LIM_V  0x100000
#define LOOP_LIM    (LOOP_LIM_V*32)

static u_char have_sum;
static DCC_SUM sum;


static int save_cksum(DCC_EMSG, DCC_WF *, DCC_CK_TYPES, DCC_SUM, DCC_TGTS);


static void NRATTRIB
usage(void)
{
	dcc_logbad(EX_USAGE, "usage: [v]"
		   " [-B ip-begin] [-E ip-end] -C 'h1 h2 h3 h4'\n");
}



int NRATTRIB
main(int argc, char **argv)
{
	in_addr_t begin, end;
	struct in6_addr addr6;
	struct in_addr addr4;
	int loop_count;
	DCC_SUM test_sum;
	char buf[80];
	int error, i;

	dcc_syslog_init(0, argv[0], 0);
	dup2(1, 2);			/* put error messages in context */

	begin = 0;
	end = -1;

	while ((i = getopt(argc, argv, "vB:E:C:")) != EOF) {
		switch (i) {
		case 'v':
			++verbose;
			break;

		case 'B':
			dcc_host_lock();
			if (!dcc_get_host(optarg, 0, &error))
				dcc_logbad(EX_USAGE, "\"-B %s\": %s",
					   optarg, DCC_HSTRERROR(error));
			begin = ntohl(dcc_hostaddrs[0].ipv4.sin_addr.s_addr);
			if (begin > end)
				dcc_logbad(EX_USAGE, "beginning after end");
			dcc_host_unlock();
			break;

		case 'E':
			dcc_host_lock();
			if (!dcc_get_host(optarg, 0, &error))
				dcc_logbad(EX_USAGE, "\"-E %s\": %s",
					   optarg, DCC_HSTRERROR(error));
			end = ntohl(dcc_hostaddrs[0].ipv4.sin_addr.s_addr);
			if (begin > end)
				dcc_logbad(EX_USAGE, "end before beginning");
			dcc_host_unlock();
			break;

		case 'C':
			if (0 >= dcc_parse_hex_ck(dcc_emsg, 0,
						  DCC_XHDR_TYPE_IP, DCC_CK_IP,
						  optarg, 0, save_cksum))
				dcc_logbad(EX_USAGE, "%s", dcc_emsg);
			have_sum = 1;
			break;

		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 0 || !have_sum)
		usage();

	dcc_clnt_unthread_init();

	addr4.s_addr = htonl(begin);
	dcc_ipv4toipv6(&addr6, addr4);
	if (verbose)
		loop_count = LOOP_LIM_V - begin % LOOP_LIM_V;
	else
		loop_count = LOOP_LIM - begin % LOOP_LIM;
	while (begin <= end) {
		dcc_ck_ipv6(test_sum, &addr6);
		if (!memcmp(test_sum, sum, sizeof(test_sum))) {
			printf("found %s\n",
			       dcc_trim_ffff(dcc_ipv6tostr(buf, sizeof(buf),
							&addr6)));
			break;
		}

		if (--loop_count <= 0) {
			printf("    %s\n",
			       dcc_trim_ffff(dcc_ipv6tostr(buf, sizeof(buf),
							&addr6)));
			loop_count = verbose ? LOOP_LIM_V : LOOP_LIM;
		}

		if (begin == 0xffffffff)
			break;
		++begin;
		addr6.s6_addr32[3] = htonl(begin);
	}

	exit(EX_OK);
}



static int
save_cksum(DCC_EMSG emsg UATTRIB, DCC_WF *wf UATTRIB,
	   DCC_CK_TYPES type UATTRIB, DCC_SUM new_sum, DCC_TGTS tgts UATTRIB)
{
	memcpy(sum, new_sum, sizeof(DCC_SUM));
	return 1;
}
